-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2020 modifications:  Mark D. Aagaard, Nusa Zidaric

-- Mark D. Aagaard
-- Riham AlTawy
-- Guang Gong
-- Kalikinkar Mandal
-- Raghvendra Rohit
-- Marat Sattarov
-- Nusa Zidaric
-- http://uwaterloo.ca/communications-security-lab/lwc/ace


-- This is a human-readable summary of (and not a substitute for) the license.
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.ace_pkg.all;
use work.util.all;
use work.ctrl_pkg.all;

entity ctrl is
  port
    ( clk
    , reset         	: in  std_logic
    ; pre_mode        : in  mode_ty
    ; pre_dom_sep     : in  dom_sep_ty
    ; pre_valid     	: in  std_logic
    ; post_ready      : in  std_logic
    ; pre_padding     : in  full_pad_ty
    ; o_control     	: out dp_op_ty
    ; o_new_ctrl 	  	: out en_ty
    ; o_pad_ctrl	  	: out padxor_ty
    ; o_oh 				    : out onehot_ty
    ; o_is_hash		    : out std_logic
    ; post_valid      : out std_logic
    ; pre_ready       : out std_logic
   );
end entity;

architecture rtl of ctrl is

  signal state,  state_next	: state_ty;
  signal count  					  : count_ty;
  signal onehot  					  : onehot_ty;

  signal new_ctrl 	 : en_ty;
  signal pad_ctrl 	 : padxor_ty;
  signal control 		 : dp_op_ty ;

  signal exit_l_load, exit_p_load, exit_p_init
  		,  exit_p_aed, exit_p_final
  		,  exit_p_abs, exit_p_sqz : 		 std_logic;

  signal rollover, counting, shifting, fullpad, newbit
  		,  valid_ed, valid_sqz, valid_t :	std_logic;

  signal m_load, m_hash, ds_aedf, ds_af
      ,  q_load, q_init, q_aedf , q_ed, q_tag, q_hash, q_sqz
		  ,  r_load, r_init, r_ed   , r_af, r_hash
		  ,  v_ed  , v_tag , v_sqz :  		  std_logic;

signal  condition_next_i_init ,	condition_next_i_aedf
	  	, condition_next_i_tag  , condition_next_i_hash
      , condition_next_i_sqz  , condition_next_l_load
      , condition_next_t_tag 	,	condition_next_p_load
      , condition_next_p_init	, condition_next_p_aed
      , condition_next_p_final, condition_next_p_abs
      , condition_next_p_sqz 	:       std_logic;

  begin

  ------------------------------------------------------------
  -- environment control
  ------------------------------------------------------------
  pre_ready 	<= '1'  when   x2b(  r_load or r_init or r_af or r_ed or r_hash )
                      else	'0';
  post_valid 	<= valid_ed when x2b( v_ed ) 	else
  						   valid_sqz when x2b( v_sqz ) else
   					     valid_t;
   valid_ed 	<= not fullpad ;
   valid_sqz 	<= '1';
   valid_t  	<= '1'  when x2b(v_tag) else 	'0';

  ------------------------------------------------------------
  -- datapath control
  ------------------------------------------------------------

  o_is_hash <= pre_mode(1);
  o_oh <= onehot;

  o_control(en_load_idx) 			 <= 	control(en_load_idx);
  o_control(sel_load_perm_idx) <= 	control(sel_load_perm_idx);

  o_control(lfsr_c_en_idx)			<=
                  (control(lfsr_c_en_idx) and is_perm_st (state) ) or rollover;
  o_control(lfsr_c_reset_idx) 	<=
                        (is_perm_r(count) or reset) when x2b(is_perm_st(state))
                         else '1' when x2b(is_l_load_st(state)) else     reset;

  o_control(sel_input_idx) <= 	control(sel_input_idx);
  o_control(sel_out1_idx)  <= 	control(sel_out1_idx);
  o_control(sel_out0_idx)  <=
                        control(sel_out0_idx) and onehot(4) when x2b(v_tag) else
  											control(sel_out0_idx)	;
  o_control(en_out_idx)    <=
                        control(en_out_idx) when x2b(v_tag or ( q_sqz)) else
  											control(en_out_idx) and not fullpad;


   o_control(sel_round_step_idx) <= 	control(sel_round_step_idx) and
                                      count(0) and count(1) and count(2);

  u_ce : entity work.ctrl_pad_to_en
                port map (  pre_padding(2 downto 0) ,  new_ctrl ,   pad_ctrl  );

  fullpad 		<= pre_padding(3);
  o_new_ctrl 	<= new_ctrl when
                 x2b( v_ed and is_dec_mode(pre_mode) and not fullpad )
                 else (others => newbit);
  o_pad_ctrl 	<= pad_ctrl when
                 x2b( v_ed and is_dec_mode(pre_mode) and not fullpad )
                 else (others => '0');

  newbit    <= '1' when
                x2b(((q_init or q_aedf) and pre_valid)
                      or v_ed or r_af or (m_hash and is_Habs_mode(pre_mode)))
                else '0';

  control 	<=
     dp_op_1 		when x2b( q_load  and  pre_valid)  										      else
  	 dp_op_2 		when x2b( is_perm_st(state) and not is_perm_done(count))		else
  	 dp_op_3 		when x2b(	(q_init and pre_valid) or	ds_af or
                          (m_hash and is_Habs_mode(pre_mode)) )					    else
  	 dp_op_4 		when x2b( q_sqz and post_ready )					                  else
  	 dp_op_5 		when x2b( (v_ed and is_enc_mode(pre_mode)) or
                          ( v_ed and is_dec_mode(pre_mode) and fullpad ))  	else
 		 dp_op_6		when x2b( v_ed and is_dec_mode(pre_mode) and not fullpad ) 	else
		 dp_op_7 		when x2b( v_tag	) 															            else
		 (others => '0');

  ------------------------------------------------------------
  -- transient states
  ------------------------------------------------------------

	  exit_l_load  <= is_l_load_st(state ) and
                  ((is_ae_mode(pre_mode) and is_load_done( onehot )) or
                    is_hash_mode(pre_mode));

	  exit_p_load 	<= is_p_load_st(state)    and is_perm_done( count );
	  exit_p_init   <= is_p_init_st(state) 	  and is_perm_done( count );
	  exit_p_aed  	<= is_p_aed_st(state) 	  and is_perm_done( count );
	  exit_p_final  <= is_p_final_st(state)   and is_perm_done( count );
	  exit_p_abs  	<= is_p_absorb_st(state) 	and is_perm_done( count );
	  exit_p_sqz  	<= is_p_squeeze_st(state) and is_perm_done( count );

    r_load <= is_i_load_st( state ) or											   			-- from idle
            ( is_l_load_st( state ) and is_ae_mode(pre_mode) and
          not is_load_done( onehot ) );                             -- loop back

    q_load <= r_load;
    m_load <=  exit_p_load;

    r_init <= is_i_init_st( state ) 		or 								          -- from idle
		 	       ( m_load and is_ae_mode(pre_mode))		or 								-- first in
  	  		   ( exit_p_init and not is_init_done(onehot) );   				-- loop back
    q_init <= r_init;

    q_aedf <=  is_i_aedf_st( state ) 				              or  -- from idle/stall
			  		  ( exit_p_init  and is_init_done(onehot) )  	or	-- first in
           	 	( exit_p_aed )  							or	 					    -- loop back 1 (from pAED)
            	( exit_p_final and not is_final_done(onehot) );	-- loop back 2 (from pFinal)

    ds_aedf <= q_aedf and pre_valid;
    q_ed  	<= ds_aedf and     is_ed_ds( pre_dom_sep );
    r_af  	<= ds_aedf and not is_ed_ds( pre_dom_sep );

    r_ed  	<= q_ed and post_ready;
    v_ed 	  <= r_ed;
    ds_af  	<= r_af;

    q_tag <= is_i_tag_st( state )                     or	  				-- from idle
     			  ( exit_p_final and is_final_done(onehot)) or 						-- first in
     			  ( is_t_tag_st(state) and not is_tag_done(onehot));		  -- loop back

    v_tag <= q_tag and post_ready;

    r_hash  <= is_i_hash_st( state ) 		              or            -- from idle
		 	        ( m_load and is_hash_mode(pre_mode))		or 	     		  -- first in
		 	        ( exit_p_abs ); 			  								 	            -- loop back

	 q_hash <= r_hash;
	 m_hash <= q_hash and pre_valid;

	 v_sqz  <= is_i_squeeze_st( state ) 		          or 							-- from idle
		 	      ( m_hash and is_Hsqz_mode(pre_mode))		or 				      -- first in
    			  ( exit_p_sqz and not is_squeeze_done(onehot));	        -- loop back

    q_sqz <= v_sqz;

  ------------------------------------------------------------
  -- next state logic
  ------------------------------------------------------------
    counting <= 	'1' when x2b( is_perm_st(state)) 	else '0';
    rollover <=  (is_perm_st (state) 	and is_perm_done( count ));
    shifting <=  (q_load and pre_valid ) or  (q_init and pre_valid ) or
                 ( ds_af and is_if_ds( pre_dom_sep ) ) or (v_tag)    or
                 ( q_sqz and post_ready );


    condition_next_l_load 	<=   q_load and pre_valid; 				-- wait for parcel
    condition_next_p_load 	<=   exit_l_load or								-- start perm
			( is_p_load_st(state) and not is_perm_done(count));     -- continue perm
    condition_next_i_init 	<=   q_init and not pre_valid;	  -- wait for parcel
    condition_next_p_init 	<= ( q_init and pre_valid ) or 		-- start perm
    	( is_p_init_st(state) and not is_perm_done(count));     -- continue perm
    condition_next_i_aedf 	<= ( q_aedf and not pre_valid ) or-- wait for parcel
    								( q_ed   and not post_ready )           or-- stall
										( is_i_aedf_st(state) and not  post_ready);-- still stalling
	 condition_next_p_aed  	  <=
      ( ds_aedf and is_ad_ds( pre_dom_sep ))   or	-- start perm: parcel=AD
	 		  v_ed                                   or	-- start perm: parcel=ptxt/ctxt
	  	( is_p_aed_st(state) and not is_perm_done( count )) ; 	-- continue perm
    condition_next_p_final	<=
      ( ds_aedf and is_if_ds( pre_dom_sep ) )  or -- start perm: parcel=K0/K1
			( is_p_final_st(state) and not is_perm_done( count ));	-- continue perm
    condition_next_t_tag  	<=   q_tag  and post_ready;				-- next tag chunk
    condition_next_i_tag 	  <=   q_tag  and not post_ready;		-- stall
    condition_next_i_hash   <=   q_hash and not pre_valid;		-- wait for parcel
    condition_next_p_abs    <=
      ( m_hash and is_Habs_mode(pre_mode))  or                  -- start perm
    	( is_p_absorb_st(state) and not is_perm_done(count));   	-- continue perm
    condition_next_i_sqz    <=    q_sqz and not post_ready;		  -- stall
    condition_next_p_sqz    <=  ( q_sqz and post_ready)  or			-- start perm
    	( is_p_squeeze_st(state) and not is_perm_done(count) and
        not is_squeeze_done(onehot));                           -- continue perm


          state_next <= l_load_st    when x2b( condition_next_l_load  ) else
            						p_load_st    when x2b( condition_next_p_load  ) else
             					  i_init_st    when x2b( condition_next_i_init  )	else
             						p_init_st    when x2b( condition_next_p_init  )	else
             						i_aedf_st    when x2b( condition_next_i_aedf  )	else
             						p_aed_st     when x2b( condition_next_p_aed   )	else
             						p_final_st   when x2b( condition_next_p_final )	else
             					  t_tag_st     when x2b( condition_next_t_tag   )	else
             						i_tag_st     when x2b( condition_next_i_tag   )	else
										    i_hash_st    when x2b( condition_next_i_hash  )	else
										    p_absorb_st  when x2b( condition_next_p_abs   )	else
										    i_squeeze_st when x2b( condition_next_i_sqz   )	else
										    p_squeeze_st when x2b( condition_next_p_sqz   )	else
    									  i_load_st;

  ------------------------------------------------------------
  -- state update
  ------------------------------------------------------------

  state_update: process(clk) begin
    if rising_edge(clk) then
      if x2b( reset ) then
        state  	<= i_load_st;
       else
        state  	<= state_next;
       end if;
    end if;
  end process;

  count_update: process(clk) begin
    if rising_edge(clk) then
       if x2b( reset ) then
       	count  	<= ( others => '0');
       else
        if 	x2b(rollover) then
       		count  	<= ( 0 => '1', others => '0'); 
       	elsif x2b(counting) then
      		count <= count + 1;
        	end if;
       end if;
    end if;
  end process;

  onehot_update: process(clk) begin
    if rising_edge(clk) then
      if x2b( reset ) then
        onehot <= (0 => '1', others => '0');
      elsif x2b(shifting) then
        onehot <= onehot_rotate( onehot );
      end if;
    end if;
  end process;

end architecture;
